package exercises

object Exercise3 extends App {
  sealed abstract class Tree {
    def isMirrorOf(other: Tree): Boolean
    def isSymmetric(): Boolean
    def computeDepth(): Int = {
      this match {
        case Empty => 1
        case Node(left, _, right) => 1 + Math.max(left.computeDepth(), right.computeDepth())
      }
    }
    def traverseBreadthFirst(): List[Int] = {
      val depth: Int = computeDepth()
      // Construct list of levels
      def helper(tree: Tree, curDepth: Int = 0): List[List[Int]] = {
        tree match {
          // Add empty levels for consistent indices
          case Empty => List.fill(depth - curDepth)(List.empty[Int])
          case Node(left, elem, right) => {
            val leftList: List[List[Int]] = helper(left, curDepth + 1)
            val rightList: List[List[Int]] = helper(right, curDepth + 1)
            val res: List[List[Int]] = (0 until depth - curDepth - 1).map(i => {
              leftList(i) ::: rightList(i)
            }).toList

            // Add this level
            List(elem)::res
          }
        }
      }
      helper(this).flatten
    }
  }

  case class Node(left: Tree, elem: Int, right: Tree) extends Tree {
    def isMirrorOf(other: Tree): Boolean = {
      other match {
        case Node(left2, _, right2) => (left isMirrorOf right2) && (right isMirrorOf left2)
        case _ => false
      }
    }
    def isSymmetric: Boolean = left isMirrorOf right
  }

  case object Empty extends Tree {
    def isMirrorOf(other: Tree): Boolean = {
      other match {
        case Empty => true
        case _ => false
      }
    }
    def isSymmetric: Boolean = true
  }
}